"""
Collect text from up to N pages under a root URL (same host), limited by max depth,
and save results to a binary file (gzip-compressed pickle).

Usage:
  python3 -m bluer_agent.crawl.collect \
    --root https://badkoobeh.com/ \
    --page-count 25 \
    --max-depth 2 \
    --out site_text.pkl.gz
"""

from __future__ import annotations


import random
import re
import signal
import time
from collections import deque
from contextlib import suppress
from typing import Deque, Dict, Iterable, Optional, Set
from urllib.parse import urljoin, urldefrag, urlparse
import requests
from bs4 import BeautifulSoup

from bluer_options.logger.config import shorten_text

from bluer_agent.crawl.classes import CrawlItem, RetryPolicy
from bluer_agent.logger import logger


class SiteTextCollector:
    def __init__(
        self,
        root_url: str,
        *,
        user_agent: str = "Mozilla/5.0 (compatible; SiteTextCollector/1.1)",
        accept_language: str = "fa,en;q=0.8",
        retry: RetryPolicy = RetryPolicy(),
    ):
        self.root_url = self._normalize_root(root_url)
        self.retry = retry

        self.session = requests.Session()
        self.session.headers.update(
            {
                "User-Agent": user_agent,
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
                "Accept-Language": accept_language,
            }
        )

        # Ctrl+C / SIGINT handling
        self._stop_requested: bool = False
        self._old_sigint_handler = None

    def _on_sigint(self, signum, frame) -> None:
        # First Ctrl+C: request a graceful stop
        if not self._stop_requested:
            self._stop_requested = True
            logger.warning(
                "Ctrl+C received: stopping after current step, will save partial results..."
            )
        else:
            # Second Ctrl+C: hard exit
            logger.warning("Ctrl+C received again: exiting immediately.")
            raise KeyboardInterrupt

    @staticmethod
    def _normalize_url(url: str) -> str:
        url, _ = urldefrag(url)  # remove #fragment
        return url.strip()

    @staticmethod
    def _url_key(url: str) -> str:
        """
        Canonical key used for de-duping.
        Treat https://badkoobeh.com and https://badkoobeh.com/ as identical.
        Only collapses the trailing slash for the *root path*.
        """
        url = SiteTextCollector._normalize_url(url)
        p = urlparse(url)

        # Only normalize the root path "/" (or empty) to no trailing slash.
        if p.scheme and p.netloc and (p.path in ("", "/")) and not p.query:
            return f"{p.scheme}://{p.netloc}"
        return url

    def _normalize_root(self, root_url: str) -> str:
        root = self._normalize_url(root_url)
        # Make urljoin predictable for relative paths:
        if not root.endswith("/"):
            root += "/"
        return root

    def _same_site(self, url: str) -> bool:
        u = urlparse(url)
        r = urlparse(self.root_url)
        if not (u.scheme and u.netloc):
            return False
        if u.scheme not in ("http", "https"):
            return False
        # strict match (scheme + host + optional port)
        return (u.scheme, u.netloc) == (r.scheme, r.netloc)

    @staticmethod
    def _is_probably_html(resp: requests.Response) -> bool:
        ctype = (resp.headers.get("Content-Type") or "").lower()
        return (
            "text/html" in ctype
            or "application/xhtml+xml" in ctype
            or ctype.startswith("text/")
        )

    def _fetch_with_retries(self, url: str) -> Optional[requests.Response]:
        for attempt in range(1, self.retry.max_retries + 1):
            if self._stop_requested:
                return None

            try:
                resp = self.session.get(
                    url,
                    timeout=self.retry.timeout_s,
                    allow_redirects=True,
                )
                # Retryable server / throttling responses
                if resp.status_code in (429, 500, 502, 503, 504):
                    raise requests.HTTPError(
                        f"Retryable HTTP {resp.status_code}", response=resp
                    )
                return resp
            except KeyboardInterrupt:
                # Respect immediate interrupts
                raise
            except Exception as e:
                logger.warning(e)

                if attempt == self.retry.max_retries:
                    break

                sleep_s = self.retry.backoff_base_s * (2 ** (attempt - 1))
                sleep_s += random.uniform(0, self.retry.backoff_jitter_s)
                time.sleep(sleep_s)

        return None

    @staticmethod
    def _extract_text(html: str) -> str:
        soup = BeautifulSoup(html, "html.parser")

        # Drop non-content / noisy areas
        for tag in soup(
            [
                "script",
                "style",
                "noscript",
                "svg",
                "canvas",
                "iframe",
                "header",
                "footer",
                "nav",
            ]
        ):
            tag.decompose()

        main = soup.find("main") or soup.find("article") or soup.body or soup

        text = main.get_text(separator="\n", strip=True)

        # Normalize: remove empty lines and compress whitespace
        lines = [ln.strip() for ln in text.splitlines()]
        lines = [ln for ln in lines if ln]
        text = "\n".join(lines)
        text = re.sub(r"[ \t]{2,}", " ", text)

        return text.strip()

    def _extract_links(self, html: str, base_url: str) -> Iterable[str]:
        soup = BeautifulSoup(html, "html.parser")

        for a in soup.find_all("a", href=True):
            if self._stop_requested:
                break

            href = (a.get("href") or "").strip()
            if not href or href.startswith(("mailto:", "tel:", "javascript:")):
                continue

            abs_url = urljoin(base_url, href)
            abs_url = self._normalize_url(abs_url)

            if not self._same_site(abs_url):
                continue

            # Skip obvious non-page assets
            path = urlparse(abs_url).path.lower()
            if path.endswith(
                (
                    ".pdf",
                    ".jpg",
                    ".jpeg",
                    ".png",
                    ".webp",
                    ".gif",
                    ".zip",
                    ".rar",
                    ".mp3",
                    ".mp4",
                    ".avi",
                    ".mov",
                )
            ):
                continue

            yield abs_url

    def collect(self, *, page_count: int, max_depth: int) -> Dict[str, str]:
        # install SIGINT handler for graceful Ctrl+C
        self._stop_requested = False
        self._old_sigint_handler = signal.getsignal(signal.SIGINT)
        signal.signal(signal.SIGINT, self._on_sigint)

        queue: Deque[CrawlItem] = deque([CrawlItem(self.root_url, 0)])
        visited: Set[str] = set()
        enqueued: Set[str] = {self._url_key(self.root_url)}  # de-dupe the queue
        results: Dict[str, str] = {}

        try:
            while queue and len(results) < page_count and not self._stop_requested:
                logger.info(
                    "fetched {} / {} page(s) - {} queued...".format(
                        len(results),
                        page_count,
                        len(queue),
                    )
                )

                item = queue.popleft()
                item_key = self._url_key(item.url)
                enqueued.discard(item_key)  # no longer in queue

                if item_key in visited:
                    continue
                visited.add(item_key)

                resp = self._fetch_with_retries(item.url)
                if self._stop_requested:
                    break
                if resp is None:
                    continue

                if not self._is_probably_html(resp):
                    continue

                # Helps Persian text when server mislabels encoding
                resp.encoding = resp.apparent_encoding or resp.encoding
                html = resp.text or ""

                text = self._extract_text(html)
                if text:
                    logger.info(
                        "📜 += {}: {}".format(
                            item.url,
                            shorten_text(text.replace("\n", " ")),
                        )
                    )
                    results[item_key] = text

                if item.depth < max_depth and not self._stop_requested:
                    for link in self._extract_links(html, base_url=item.url):
                        if self._stop_requested:
                            break
                        link_key = self._url_key(link)
                        if link_key not in visited and link_key not in enqueued:
                            logger.info(f"🔗 += {link}")
                            queue.append(CrawlItem(link, item.depth + 1))
                            enqueued.add(link_key)

                if self.retry.delay_between_requests_s > 0 and not self._stop_requested:
                    time.sleep(self.retry.delay_between_requests_s)

            return results

        finally:
            # restore original SIGINT handler + close HTTP session
            with suppress(Exception):
                signal.signal(signal.SIGINT, self._old_sigint_handler)
            with suppress(Exception):
                self.session.close()



from typing import List

from bluer_options.terminal import show_usage, xtra


def help_install(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@audio",
            "install",
        ],
        "install.",
        mono=mono,
    )


def help_play(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("download,filename=<audio.wav>", mono=mono)

    return show_usage(
        [
            "@audio",
            "play",
            f"[{options}]",
            "[.|<object-name>]",
        ],
        "play <object-name>/<audio.wav>.",
        mono=mono,
    )


record_args = [
    "[--channels <1>]",
    "[--crop_silence <0>]",
    "[--length <30>]",
    "[--rate <16000>]",
]


def help_record(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "".join(
        [
            xtra("filename=<audio.wav>,", mono=mono),
            "play",
            xtra(",upload", mono=mono),
        ]
    )

    return show_usage(
        [
            "@audio",
            "record",
            f"[{options}]",
            "[-|<object-name>]",
        ]
        + record_args,
        "record <object-name>/<audio.wav>.",
        mono=mono,
    )


help_functions = {
    "install": help_install,
    "play": help_play,
    "record": help_record,
}



from bluer_options.env import load_config, load_env, get_env

load_env(__name__)
load_config(__name__)

BLUER_AGENT_API_KEY = get_env("BLUER_AGENT_API_KEY")

BLUER_AGENT_AUDIO_TEST_OBJECT = get_env("BLUER_AGENT_AUDIO_TEST_OBJECT")

BLUER_AGENT_CHAT_ENDPOINT = get_env("BLUER_AGENT_CHAT_ENDPOINT")
BLUER_AGENT_CHAT_MAX_TOKENS = get_env("BLUER_AGENT_CHAT_MAX_TOKENS", 3000)
BLUER_AGENT_CHAT_MODEL_NAME = get_env("BLUER_AGENT_CHAT_MODEL_NAME")
BLUER_AGENT_CHAT_TEMPERATURE = get_env("BLUER_AGENT_CHAT_TEMPERATURE", 0.7)
BLUER_AGENT_CHAT_TEST_OBJECT = get_env("BLUER_AGENT_CHAT_TEST_OBJECT")
BLUER_AGENT_CHAT_TIMEOUT = get_env("BLUER_AGENT_CHAT_TIMEOUT", 60)

BLUER_AGENT_MACHINE_USER_NAME = get_env("BLUER_AGENT_MACHINE_USER_NAME")

BLUER_AGENT_RAG_TEST_CORPUS = get_env("BLUER_AGENT_RAG_TEST_CORPUS")

BLUER_AGENT_TRANSCRIPTION_ENDPOINT = get_env("BLUER_AGENT_TRANSCRIPTION_ENDPOINT")
BLUER_AGENT_TRANSCRIPTION_MODEL_NAME = get_env("BLUER_AGENT_TRANSCRIPTION_MODEL_NAME")
BLUER_AGENT_TRANSCRIPTION_RETRIAL = get_env("BLUER_AGENT_TRANSCRIPTION_RETRIAL", 5)
BLUER_AGENT_TRANSCRIPTION_TEST_OBJECT = get_env("BLUER_AGENT_TRANSCRIPTION_TEST_OBJECT")


# Thresholds: percent of full scale. Try 1%..5%.
# how loud to begin
BLUER_AGENT_AUDIO_LISTEN_START_THRESHOLD = "2%"
# how quiet to consider silence
BLUER_AGENT_AUDIO_LISTEN_STOP_THRESHOLD = "2%"

# durations:
# need this long above threshold to start (sec)
BLUER_AGENT_AUDIO_LISTEN_START_HOLD = "0.10"
# stop after this long below threshold (sec)
BLUER_AGENT_AUDIO_LISTEN_STOP_HOLD = "1.20"



NAME = "bluer_agent"

ICON = "👾"

DESCRIPTION = f"{ICON} an AI agent."

VERSION = "5.120.1"

REPO_NAME = "bluer-agent"

MARQUEE = "https://github.com/kamangir/assets2/raw/main/bluer-agent/icon.jpg?raw=true"

ALIAS = "@agent"


def fullname() -> str:
    return f"{NAME}-{VERSION}"



import argparse

from blueness import module
from blueness.argparse.generic import sys_exit

from bluer_agent import NAME
from bluer_agent.chat.functions import chat
from bluer_agent.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="validate",
)
args = parser.parse_args()

success = False
if args.task == "validate":
    success, _ = chat(
        messages=[
            {
                "role": "user",
                "content": "چرا آسمان آبی است؟",
            }
        ]
    )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from typing import List

from bluer_options.terminal import show_usage, xtra

from bluer_agent.help.audio import record_args


def help(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "".join(
        [
            xtra("download,filename=<filename.wav>,install,", mono=mono),
            "play",
            xtra(",upload", mono=mono),
        ]
    )

    args = sorted(
        [
            "[--language en|fa]",
            "[--record 0]",
        ]
        + record_args
    )

    return show_usage(
        [
            "@transcribe",
            f"[{options}]",
            "[-|<object-name>]",
        ]
        + args,
        "validate transcription.",
        mono=mono,
    )



from blueness import module
from bluer_options.logger.config import log_list
from bluer_objects.metadata import get_from_object

from bluer_agent import NAME
from bluer_agent.logger import logger


NAME = module.name(__file__, NAME)


def build(object_name: str) -> bool:
    logger.info(f"{NAME}.build({object_name})")

    corpus_roots = get_from_object(object_name, "corpus", [])
    log_list(logger, "loaded", corpus_roots, "root(s)")

    logger.info("🪄")

    return True



#!/usr/bin/env python3
from typing import Dict
import argparse

from blueness import module
from bluer_objects import objects
from bluer_objects.metadata import post_to_object, get_from_object

from bluer_agent import NAME
from bluer_agent.crawl.classes import RetryPolicy
from bluer_agent.crawl.functions import url_to_filename
from bluer_agent.crawl.collector import SiteTextCollector
from bluer_agent.crawl import file
from bluer_agent.logger import logger


NAME = module.name(__file__, NAME)


class CollectionProperties:
    def __init__(
        self,
        timeout: float = 15.0,
        max_retries: int = 4,
        backoff_base: float = 0.7,
        backoff_jitter: float = 0.4,
        delay: float = 0.2,
    ):
        self.timeout: float = timeout
        self.max_retries: int = max_retries
        self.backoff_base: float = backoff_base
        self.backoff_jitter: float = backoff_jitter
        self.delay: float = delay


def collect(
    root: str,
    page_count: int = 25,
    max_depth: int = 2,
    object_name: str = "",
    out: str = "site_text.pkl.gz",
    properties: CollectionProperties = CollectionProperties(),
) -> bool:
    logger.info(f"{NAME}.collect({root})...")

    retry = RetryPolicy(
        max_retries=properties.max_retries,
        timeout_s=properties.timeout,
        backoff_base_s=properties.backoff_base,
        backoff_jitter_s=properties.backoff_jitter,
        delay_between_requests_s=properties.delay,
    )

    collector = SiteTextCollector(root, retry=retry)

    results: Dict[str, str] = {}
    interrupted = False
    try:
        results = collector.collect(
            page_count=page_count,
            max_depth=max_depth,
        )
    except KeyboardInterrupt:
        interrupted = True
        logger.warning("Ctrl+C, saving partial results...")

    if not file.save(
        results,
        (
            objects.path_of(
                object_name=object_name,
                filename="{}.pkl.gz".format(url_to_filename(root)),
            )
            if out == "auto"
            else out
        ),
    ):
        return False

    if object_name:
        collection_log = get_from_object(
            object_name,
            "crawl_collection",
            {},
        )

        collection_log[root] = list(results.keys())

        if not post_to_object(
            object_name,
            "crawl_collection",
            collection_log,
        ):
            return False

    if interrupted:
        # raise SystemExit(130)
        logger.error("interrupted...")
        return False

    return True



docs = [
    {
        "path": "../docs/crawl",
    },
    {
        "path": "../docs/crawl/one.md",
    },
    {
        "path": "../docs/crawl/two.md",
    },
    {
        "path": "../docs/crawl/three.md",
    },
]



from typing import Tuple
import json

from blueness import module
from bluer_options import string
from bluer_objects import file, objects
from bluer_objects.host import shell
from bluer_objects.metadata import post_to_object

from bluer_agent import NAME
from bluer_agent import env
from bluer_agent.audio.properties import AudioProperties
from bluer_agent.audio.record import record as record_audio
from bluer_agent.logger import logger


NAME = module.name(__file__, NAME)


def transcribe(
    object_name: str,
    filename: str = "",
    language: str = "fa",
    record: bool = False,
    properties: AudioProperties = AudioProperties(),
    post_metadata: bool = False,
) -> Tuple[bool, str]:
    if not filename:
        filename = "{}.wav".format(string.timestamp())

    if record and not record_audio(
        object_name=object_name,
        filename=filename,
        properties=properties,
    ):
        return False, ""

    logger.info(
        "{}.transcribe({}/{}) [{}]".format(
            NAME,
            object_name,
            file.name_and_extension(filename),
            language,
        )
    )

    # https://docs.arvancloud.ir/fa/aiaas/api-usage
    attempt = 1
    success: bool = False
    text: str = ""
    while attempt <= env.BLUER_AGENT_TRANSCRIPTION_RETRIAL:
        if attempt >= 2:
            logger.info(
                f"attempt {attempt} / {env.BLUER_AGENT_TRANSCRIPTION_RETRIAL}..."
            )

        command = [
            "curl",
            "--fail",
            f'--location "{env.BLUER_AGENT_TRANSCRIPTION_ENDPOINT}/audio/transcriptions"',
            f'--header "Authorization: apikey {env.BLUER_AGENT_API_KEY}"',
            f'--form "model={env.BLUER_AGENT_TRANSCRIPTION_MODEL_NAME}"',
            '--form "file=@{}"'.format(
                objects.path_of(
                    object_name=object_name,
                    filename=filename,
                )
            ),
            f'--form "language={language}"',
        ]

        success, output = shell(
            command,
            return_output=True,
            clean_after=True,
        )
        if success:
            if not output:
                logger.warning("silence detected.")
                text = ""
                break

            try:
                output_dict = json.loads(" ".join(output))

                assert isinstance(
                    output_dict, dict
                ), f"dict expected, received {output_dict.__class__.__name__}"

                assert "text" in output_dict, f'"text" not found in {output_dict}'

                text = output_dict["text"]

                break
            except Exception as e:
                logger.warning(f"bad output: {output}")
                logger.warning(e)

        logger.warning(f"transcription failed (attempt {attempt}).")
        attempt += 1

    if not success:
        logger.error(
            f"reached maximum retry limit ({env.BLUER_AGENT_TRANSCRIPTION_RETRIAL})."
        )
        return False, ""

    logger.info(text)

    return (
        not post_metadata
        or post_to_object(
            object_name,
            file.name(filename),
            text,
        ),
        text,
    )



import gzip
import pickle
from typing import Dict, Tuple

from blueness import module
from bluer_options import string
from bluer_options.logger.config import log_list, shorten_text
from bluer_objects import file

from bluer_agent import NAME
from bluer_agent.logger import logger


NAME = module.name(__file__, NAME)

"""
using gzip-compressed pickle:
    - binary
    - compact
    - easy to load back in Python
"""


def save(
    results: Dict[str, str],
    filename: str,
) -> bool:
    if not results:
        logger.warning("no pages collected; nothing to save.")
        return True

    payload = {
        "format": "site_text_collector",
        "version": 1,
        "results": results,
    }
    try:
        with gzip.open(filename, "wb") as f:
            pickle.dump(payload, f, protocol=pickle.HIGHEST_PROTOCOL)
    except Exception as e:
        logger.error(e)
        return False

    logger.info(
        "{}.save: {} page(s) -> {} [{}]".format(
            NAME,
            len(results),
            filename,
            string.pretty_bytes(file.size(filename)),
        )
    )
    return True


def load(filename: str) -> Tuple[bool, Dict[str, str]]:
    try:
        with gzip.open(filename, "rb") as f:
            payload = pickle.load(f)
    except Exception as e:
        logger.error(e)
        return False, {}

    if (
        not isinstance(payload, dict)
        or payload.get("format") != "site_text_collector"
        or "results" not in payload
    ):
        logger.error("unrecognized binary format.")
        return False, {}

    results = payload["results"]

    log_list(
        logger,
        "loaded from {} [{}]".format(
            filename,
            string.pretty_bytes(file.size(filename)),
        ),
        [
            "{}: {}".format(
                key,
                shorten_text(text.replace("\n", " ")),
            )
            for key, text in results.items()
        ],
        "page(s)",
    )

    return True, results



import re


def url_to_filename(url: str) -> str:
    return re.sub(
        r"[^a-zA-Z0-9_-]",
        "_",
        url.replace("https://", "")
        .replace("http://", "")
        .replace(".", "_")
        .rstrip("/"),
    )



from bluer_objects.README.alias import list_of_aliases

from bluer_agent import NAME
from bluer_agent.README import aliases, audio, chat, crawl, rag, transcription

docs = (
    [
        {
            "path": "../..",
            "macros": {
                "aliases:::": list_of_aliases(NAME),
            },
        },
        {
            "path": "../docs",
        },
    ]
    + aliases.docs
    + audio.docs
    + chat.docs
    + crawl.docs
    + rag.docs
    + transcription.docs
)



from typing import Tuple, Any

from blueness import module

from bluer_agent import NAME
from bluer_agent.logger import logger


NAME = module.name(__file__, NAME)


def query(
    object_name: str,
    query: str,
) -> Tuple[bool, Any]:
    logger.info(f'{NAME}.query[{object_name}]("{query}")')

    logger.info("🪄")

    return True, ["void"]
